@@ -16,6 +16,7 @@ from fsspec import AbstractFileSystem
 from fsspec.spec import AbstractBufferedFile
 from fsspec.utils import infer_storage_options, tokenize
 
+
 logger = logging.getLogger(__name__)
 
 
@@ -103,9 +104,10 @@ class AzureDatalakeFileSystem(AbstractFileSystem):
         )
 
         for file in files:
-            if "type" in file and file["type"] == "DIRECTORY":
-                file["type"] = "directory"
-
+            if "type" in file:
+                file["type"] = file["type"].lower()
+            if "length" in file:
+                file["size"] = file["length"]
         return files
 
     def info(self, path, invalidate_cache=True, expected_error_code=404, **kwargs):
@@ -115,6 +117,8 @@ class AzureDatalakeFileSystem(AbstractFileSystem):
             expected_error_code=expected_error_code,
         )
         info["size"] = info["length"]
+        """Azure FS uses upper case type values but fsspec is expecting lower case"""
+        info["type"] = info["type"].lower()
         return info
 
     def _trim_filename(self, fn, **kwargs):
@@ -204,13 +208,14 @@ class AzureDatalakeFile(AzureDLFile):
         self.path = AzureDLPath(path)
         self.mode = mode
 
-    def seek(self, loc, whence=0, **kwargs):
+    def seek(self, loc: int, whence: int = 0, **kwargs):
         """ Set current file location
 
         Parameters
         ----------
         loc: int
             byte location
+
         whence: {0, 1, 2}
             from start of file, current location or end of file, resp.
         """
@@ -237,62 +242,49 @@ class AzureBlobFileSystem(AbstractFileSystem):
 
     Parameters
     ----------
-    account_name:
+    account_name: str
         The storage account name. This is used to authenticate requests
         signed with an account key and to construct the storage endpoint. It
         is required unless a connection string is given, or if a custom
         domain is used with anonymous authentication.
-    container: str,
-        The container name.  Required to create a filesystem client.
-    account_key:
+    account_key: str
         The storage account key. This is used for shared key authentication.
         If any of account key, sas token or client_id is specified, anonymous access
         will be used.
-    sas_token:
+    sas_token: str
         A shared access signature token to use to authenticate requests
         instead of the account key. If account key and sas token are both
         specified, account key will be used to sign. If any of account key, sas token
         or client_id are specified, anonymous access will be used.
-    is_emulated:
-        Whether to use the emulator. Defaults to False. If specified, will
-        override all other parameters besides connection string and request
-        session.
-    protocol:
-        The protocol to use for requests. Defaults to https.
-    endpoint_suffix:
-        The host base component of the url, minus the account name. Defaults
-        to Azure (core.windows.net). Override this to use the China cloud
-        (core.chinacloudapi.cn).
-    custom_domain:
-        The custom domain to use. This can be set in the Azure Portal. For
-        example, 'www.mydomain.com'.
-    request_session:
+    request_session: Session
         The session object to use for http requests.
-    connection_string:
+    connection_string: str
         If specified, this will override all other parameters besides
         request session. See
         http://azure.microsoft.com/en-us/documentation/articles/storage-configure-connection-string/
         for the connection string format.
-    socket_timeout:
-        If specified, this will override the default socket timeout. The timeout specified is in seconds.
+    socket_timeout: int
+        If specified, this will override the default socket timeout. The timeout specified is in
+        seconds.
         See DEFAULT_SOCKET_TIMEOUT in _constants.py for the default value.
-    token_credential:
-        A token credential used to authenticate HTTPS requests. The token value
-        should be updated before its expiration.
-    blocksize:
+    credential: TokenCredential or SAS token
+        The credentials with which to authenticate.  Optional if the account URL already has a SAS token.
+        Can include an instance of TokenCredential class from azure.identity
+    blocksize: int
         The block size to use for download/upload operations. Defaults to the value of
         ``BlockBlobService.MAX_BLOCK_SIZE``
-    client_id:
+    client_id: str
         Client ID to use when authenticating using an AD Service Principal client/secret.
-    client_secret:
+    client_secret: str
         Client secret to use when authenticating using an AD Service Principal client/secret.
-    tenant_id:
+    tenant_id: str
         Tenant ID to use when authenticating using an AD Service Principal client/secret.
 
     Examples
     --------
+    Authentication with an account_key
     >>> abfs = AzureBlobFileSystem(account_name="XXXX", account_key="XXXX", container_name="XXXX")
-    >>> adl.ls('')
+    >>> abfs.ls('')
 
     **  Sharded Parquet & csv files can be read as: **
         ------------------------------------------
@@ -301,6 +293,18 @@ class AzureBlobFileSystem(AbstractFileSystem):
 
         ddf = dd.read_parquet('abfs://container_name/folder.parquet', storage_options={
         ...    'account_name': ACCOUNT_NAME, 'account_key': ACCOUNT_KEY,})
+
+    Authentication with an Azure ServicePrincipal
+    >>> abfs = AzureBlobFileSystem(account_name="XXXX", tenant_id=TENANT_ID,
+        ...    client_id=CLIENT_ID, client_secret=CLIENT_SECRET)
+    >>> abfs.ls('')
+
+    **  Read files as: **
+        -------------
+        ddf = dd.read_csv('abfs://container_name/folder/*.csv', storage_options={
+            'account_name': ACCOUNT_NAME, 'tenant_id': TENANT_ID, 'client_id': CLIENT_ID,
+            'client_secret': CLIENT_SECRET})
+        })
     """
 
     protocol = "abfs"
@@ -313,12 +317,11 @@ class AzureBlobFileSystem(AbstractFileSystem):
         credential: str = None,
         sas_token: str = None,
         request_session=None,
-        socket_timeout=None,
-        token_credential=None,
-        blocksize=create_configuration(storage_sdk="blob").max_block_size,
-        client_id=None,
-        client_secret=None,
-        tenant_id=None,
+        socket_timeout: int = None,
+        blocksize: int = create_configuration(storage_sdk="blob").max_block_size,
+        client_id: str = None,
+        client_secret: str = None,
+        tenant_id: str = None,
     ):
         AbstractFileSystem.__init__(self)
         self.account_name = account_name
@@ -328,23 +331,35 @@ class AzureBlobFileSystem(AbstractFileSystem):
         self.sas_token = sas_token
         self.request_session = request_session
         self.socket_timeout = socket_timeout
-        self.token_credential = token_credential
         self.blocksize = blocksize
         self.client_id = client_id
         self.client_secret = client_secret
         self.tenant_id = tenant_id
 
         if (
-            self.token_credential is None
+            self.credential is None
             and self.account_key is None
             and self.sas_token is None
             and self.client_id is not None
         ):
-            self.token_credential = self._get_token_from_service_principal()
+            self.credential = self._get_credential_from_service_principal()
         self.do_connect()
 
     @classmethod
-    def _strip_protocol(cls, path):
+    def _strip_protocol(cls, path: str):
+        """
+        Remove the protocol from the input path
+
+        Parameters
+        ----------
+        path: str
+            Path to remove the protocol from
+
+        Returns
+        -------
+        str
+            Returns a path without the protocol
+        """
         logging.debug(f"_strip_protocol for {path}")
         ops = infer_storage_options(path)
 
@@ -358,21 +373,32 @@ class AzureBlobFileSystem(AbstractFileSystem):
         logging.debug(f"_strip_protocol({path}) = {ops}")
         return ops["path"]
 
-    def _get_token_from_service_principal(self):
-        from azure.common.credentials import ServicePrincipalCredentials
-        from azure.storage.common import TokenCredential
+    def _get_credential_from_service_principal(self):
+        """
+        Create a Credential for authentication.  This can include a TokenCredential
+        client_id, client_secret and tenant_id
 
-        sp_cred = ServicePrincipalCredentials(
+        Returns
+        -------
+        Credential
+        """
+        from azure.identity import ClientSecretCredential
+
+        sp_token = ClientSecretCredential(
+            tenant_id=self.tenant_id,
             client_id=self.client_id,
-            secret=self.client_secret,
-            tenant=self.tenant_id,
-            resource="https://storage.azure.com/",
+            client_secret=self.client_secret,
         )
-
-        token_cred = TokenCredential(sp_cred.token["access_token"])
-        return token_cred
+        return sp_token
 
     def do_connect(self):
+        """Connect to the BlobServiceClient, using user-specified connection details.
+        Tries credentials first, then connection string and finally account key
+
+        Raises
+        ------
+        ValueError if none of the connection details are available
+        """
         self.account_url: str = f"https://{self.account_name}.blob.core.windows.net"
         if self.credential is not None:
             self.service_client = BlobServiceClient(
@@ -386,16 +412,27 @@ class AzureBlobFileSystem(AbstractFileSystem):
             self.service_client = BlobServiceClient(
                 account_url=self.account_url, credential=self.account_key
             )
+        elif self.sas_token is not None:
+            self.service_client = BlobServiceClient(
+                account_url=self.account_url + self.sas_token, credential=None
+            )
         else:
             raise ValueError("unable to connect with provided params!!")
 
     def split_path(self, path, delimiter="/", return_container: bool = False, **kwargs):
         """
         Normalize ABFS path string into bucket and key.
+
         Parameters
         ----------
         path : string
             Input path, like `abfs://my_container/path/to/file`
+
+        delimiter: string
+            Delimiter used to split the path
+
+        return_container: bool
+
         Examples
         --------
         >>> split_path("abfs://my_container/path/to/file")
@@ -413,39 +450,6 @@ class AzureBlobFileSystem(AbstractFileSystem):
         else:
             return path.split(delimiter, 1)
 
-    # def _generate_blobs(self, *args, **kwargs):
-    #     """Follow next_marker to get ALL results."""
-    #     logging.debug("running _generate_blobs...")
-    #     blobs = self.blob_fs.list_blobs(*args, **kwargs)
-    #     yield from blobs
-    #     while blobs.next_marker:
-    #         logging.debug(f"following next_marker {blobs.next_marker}")
-    #         kwargs["marker"] = blobs.next_marker
-    #         blobs = self.blob_fs.list_blobs(*args, **kwargs)
-    #         yield from blobs
-
-    # def _matches(
-    #     self, container_name, path, as_directory=False, delimiter="/", **kwargs
-    # ):
-    #     """check if the path returns an exact match"""
-
-    #     path = path.rstrip(delimiter)
-    #     gen = self.blob_fs.list_blob_names(
-    #         container_name=container_name,
-    #         prefix=path,
-    #         delimiter=delimiter,
-    #         num_results=None,
-    #     )
-
-    #     contents = list(gen)
-    #     if not contents:
-    #         return False
-
-    #     if as_directory:
-    #         return contents[0] == path + delimiter
-    #     else:
-    #         return contents[0] == path
-
     def ls(
         self,
         path: str,
@@ -460,15 +464,27 @@ class AzureBlobFileSystem(AbstractFileSystem):
 
         Parameters
         ----------
-        path:  Path to an Azure Blob with its container name
-        detail:  If False, return a list of blob names, else a list of dictionaries with blob details
-        invalidate_cache:  Boolean
-        """
+        path: str
+            Path to an Azure Blob with its container name
 
+        detail: bool
+            If False, return a list of blob names, else a list of dictionaries with blob details
+
+        invalidate_cache:  bool
+            If True, do not use the cache
+
+        delimiter: str
+            Delimiter used to split paths
+
+        return_glob: bool
+
+        """
+        # import pdb;pdb.set_trace()
         logging.debug(f"abfs.ls() is searching for {path}")
 
+        target_path = path.lstrip("/")
         container, path = self.split_path(path)
-        if (container in ["", delimiter]) and (path in ["", delimiter]):
+        if (container in ["", ".", delimiter]) and (path in ["", delimiter]):
             # This is the case where only the containers are being returned
             logging.info(
                 "Returning a list of containers in the azure blob storage account"
@@ -524,11 +540,20 @@ class AzureBlobFileSystem(AbstractFileSystem):
                                 return self._details(blob_page)
                         else:
                             outblobs = []
+                            depth = target_path.count("/")
+                            if depth == 0:
+                                depth = 2
+                            else:
+                                depth = depth + 1
                             for blob_page in blobs:
                                 for blob in blob_page:
-                                    outblobs.append(
+                                    directory_ = (
                                         f"{blob.container}{delimiter}{blob.name}"
                                     )
+                                    dir_parts = directory_.split("/", depth)[0:depth]
+                                    directory = "/".join([d for d in dir_parts])
+                                    directory = f"{directory}/"
+                                    outblobs.append(directory)
                             return outblobs
                     elif blobs[0]["blob_type"] == "BlockBlob":
                         if detail:
@@ -556,6 +581,24 @@ class AzureBlobFileSystem(AbstractFileSystem):
                     raise FileNotFoundError
 
     def _details(self, contents, delimiter="/", return_glob: bool = False, **kwargs):
+        """
+        Return a list of dictionaries of specifying details about the contents
+
+        Parameters
+        ----------
+        contents
+
+        delimiter: str
+            Delimiter used to separate containers and files
+
+        return_glob: bool
+
+
+        Returns
+        -------
+        List of dicts
+            Returns details about the contents, such as name, size and type
+        """
         pathlist = []
         for c in contents:
             data = {}
@@ -579,7 +622,7 @@ class AzureBlobFileSystem(AbstractFileSystem):
             pathlist.append(data)
         return pathlist
 
-    def walk(self, path, maxdepth=None, **kwargs):
+    def walk(self, path: str, maxdepth=None, **kwargs):
         """ Return all files belows path
 
         List all files, recursing into subdirectories; output is iterator-style,
@@ -592,10 +635,12 @@ class AzureBlobFileSystem(AbstractFileSystem):
         ----------
         path: str
             Root to recurse into
+
         maxdepth: int
             Maximum recursion depth. None means limitless, but not recommended
             on link-based file-systems.
-        kwargs: passed to ``ls``
+
+        **kwargs are passed to ``ls``
         """
         path = self._strip_protocol(path)
         full_dirs = {}
@@ -637,12 +682,29 @@ class AzureBlobFileSystem(AbstractFileSystem):
             yield from self.walk(d, maxdepth=maxdepth, detail=detail, **kwargs)
 
     def mkdir(self, path, delimiter="/", exists_ok=False, **kwargs):
+        """
+        Create directory entry at path
+
+        Parameters
+        ----------
+        path: str
+            The path to create
+
+        delimiter: str
+            Delimiter to use when splitting the path
+
+        exists_ok: bool
+            If True, raise an exception if the directory already exists. Defaults to False
+        """
         container_name, path = self.split_path(path, delimiter=delimiter)
         if not exists_ok:
             if (container_name not in self.ls("")) and (not path):
                 # create new container
                 self.service_client.create_container(name=container_name)
-            elif (container_name in self.ls("")) and path:
+            elif (
+                container_name
+                in [container_path.split("/")[0] for container_path in self.ls("")]
+            ) and path:
                 ## attempt to create prefix
                 container_client = self.service_client.get_container_client(
                     container=container_name
@@ -658,41 +720,90 @@ class AzureBlobFileSystem(AbstractFileSystem):
                 )
                 container_client.upload_blob(name=path, data="")
 
-    def rmdir(self, path, delimiter="/", **kwargs):
+    def rmdir(self, path: str, delimiter="/", **kwargs):
+        """
+        Remove a directory, if empty
+
+        Parameters
+        ----------
+        path: str
+            Path of directory to remove
+
+        delimiter: str
+            Delimiter to use when splitting the path
+
+        """
+
         container_name, path = self.split_path(path, delimiter=delimiter)
         if (container_name + delimiter in self.ls("")) and (not path):
             # delete container
             self.service_client.delete_container(container_name)
 
-    def _rm(self, path, delimiter="/", **kwargs):
-        if self.isfile(path):
-            container_name, path = self.split_path(path, delimiter=delimiter)
-            container_client = self.service_client.get_container_client(
-                container=container_name
-            )
-            logging.debug(f"Delete blob {path} in {container_name}")
-            container_client.delete_blob(path)
-        elif self.isdir(path):
-            container_name, path = self.split_path(path, delimiter=delimiter)
-            container_client = self.service_client.get_container_client(
-                container=container_name
-            )
-            if (container_name + delimiter in self.ls("")) and (not path):
-                logging.debug(f"Delete container {container_name}")
-                container_client.delete_container(container_name)
-        else:
-            raise RuntimeError(f"cannot delete {path}")
+    def rm_file(self, path, delimiter="/", **kwargs):
+        """
+        Delete a given file
+
+        Parameters
+        ----------
+        path: str
+            Path to file to delete
+
+        delimiter: str
+            Delimiter to use when splitting the path
+        """
+        try:
+            kind = self.info(path)["type"]
+            if kind == "file":
+                container_name, path = self.split_path(path, delimiter=delimiter)
+                container_client = self.service_client.get_container_client(
+                    container=container_name
+                )
+                logging.debug(f"Delete blob {path} in {container_name}")
+                container_client.delete_blob(path)
+            elif kind == "directory":
+                container_name, path = self.split_path(path, delimiter=delimiter)
+                container_client = self.service_client.get_container_client(
+                    container=container_name
+                )
+                if (container_name + delimiter in self.ls("")) and (not path):
+                    logging.debug(f"Delete container {container_name}")
+                    container_client.delete_container()
+            else:
+                raise RuntimeError(f"Unable to delete {path}!")
+        except FileNotFoundError:
+            pass
 
     def _open(
         self,
-        path,
-        mode="rb",
-        block_size=None,
-        autocommit=True,
+        path: str,
+        mode: str = "rb",
+        block_size: int = None,
+        autocommit: bool = True,
         cache_options=None,
         **kwargs,
     ):
-        """ Open a file on the datalake, or a block blob """
+        """Open a file on the datalake, or a block blob
+
+        Parameters
+        ----------
+        path: str
+            Path to file to open
+
+        mode: str
+            What mode to open the file in - defaults to "rb"
+
+        block_size: int
+            Size per block for multi-part downloads.
+
+        autocommit: bool
+            Whether or not to write to the destination directly
+
+        cache_type: str
+            One of "readahead", "none", "mmap", "bytes", defaults to "readahead"
+            Caching policy in read mode.
+            See the definitions here:
+            https://filesystem-spec.readthedocs.io/en/latest/api.html#readbuffering
+        """
         logging.debug(f"_open:  {path}")
         return AzureBlobFile(
             fs=self,
@@ -710,15 +821,47 @@ class AzureBlobFile(AbstractBufferedFile):
 
     def __init__(
         self,
-        fs,
-        path,
-        mode="rb",
+        fs: AzureBlobFileSystem,
+        path: str,
+        mode: str = "rb",
         block_size="default",
-        autocommit=True,
-        cache_type="readahead",
-        cache_options=None,
+        autocommit: bool = True,
+        cache_type: str = "readahead",
+        cache_options: dict = None,
         **kwargs,
     ):
+        """
+        Represents a file on AzureBlobStorage that implements buffered reading and writing
+
+        Parameters
+        ----------
+        fs: AzureBlobFileSystem
+            An instance of the filesystem
+
+        path: str
+            The location of the file on the filesystem
+
+        mode: str
+            What mode to open the file in. Defaults to "rb"
+
+        block_size: int, str
+            Buffer size for reading and writing. The string "default" will use the class
+            default
+
+        autocommit: bool
+            Whether or not to write to the destination directly
+
+        cache_type: str
+            One of "readahead", "none", "mmap", "bytes", defaults to "readahead"
+            Caching policy in read mode. See the definitions in ``core``.
+
+        cache_options : dict
+            Additional options passed to the constructor for the cache specified
+            by `cache_type`.
+
+        kwargs: dict
+            Passed to AbstractBufferedFile
+        """
         container_name, blob = fs.split_path(path)
         self.container_name = container_name
         self.blob = blob
@@ -737,13 +880,24 @@ class AzureBlobFile(AbstractBufferedFile):
             **kwargs,
         )
 
-    def _fetch_range(self, start, end, **kwargs):
+    def _fetch_range(self, start: int, end: int, **kwargs):
+        """
+        Download a chunk of data specified by start and end
+
+        Parameters
+        ----------
+        start: int
+            Start byte position to download blob from
+        end: int
+            End byte position to download blob from
+        """
         blob = self.container_client.download_blob(
             blob=self.blob, offset=start, length=end
         )
         return blob.readall()
 
     def _initiate_upload(self, **kwargs):
+        """Prepare a remote file upload"""
         self.blob_client = self.container_client.get_blob_client(blob=self.blob)
         self._block_list = []
         try:
@@ -751,11 +905,21 @@ class AzureBlobFile(AbstractBufferedFile):
         except ResourceNotFoundError:
             pass
         except Exception as e:
-            raise f"Failed for {e}"
+            raise RuntimeError(f"Failed for {e}")
         else:
             return super()._initiate_upload()
 
-    def _upload_chunk(self, final=False, **kwargs):
+    def _upload_chunk(self, final: bool = False, **kwargs):
+        """
+        Write one part of a multi-block file upload
+
+        Parameters
+        ----------
+        final: bool
+            This is the last block, so should complete file, if
+            self.autocommit is True.
+
+        """
         data = self.buffer.getvalue()
         length = len(data)
         block_id = len(self._block_list)
